// Evolutionary rescue of partially clonal organisms in fluctuating environments


/************************************************************************
 Description
 ************************************************************************/
/*
 This is the code for a quantitative genetic individual-based model designed
 to study how environmental variation affects the relationship between clonality
 and evolutionary rescue seen in Orive et al. 2017
 
 This code simulates an abrupt environment change. See 'directional.cpp' for a
 model of directional environmental change
 */


/************************************************************************
 Includes
 ************************************************************************/

// standard includes
#include <iostream>
#include <vector>
#include <math.h>
#include <random>
#include <algorithm>
#include <chrono>  // for high_resolution_clock


/************************************************************************
 Random Number Generator
 ************************************************************************/
// Psuedo-random number generator.
// It uses a Mersenne Twister algorithm to generate the random numbers.
// Then transforms their distributions accordingly

typedef std::mt19937 RNG; // making typedef for the random number generator

// Uniform distribution
struct Unif {
    double i; // min
    double k; // max
    RNG rng; // random number generator object
    Unif(RNG& rng_) : rng(rng_) {}
    // contructor takes seed
    // can make seed drawn from random device if you want it different every time
    
    inline double doub(double min,double max) { // outputs double
        // takes min and max of the range
        i = min; k = max;
        std::uniform_real_distribution<double> dist(i,k);
        return dist(rng);
    }
    
    inline float flo(double min,double max) { // outputs float
        // takes min and max of the range
        i = min; k = max;
        std::uniform_real_distribution<float> dist(i,k);
        return dist(rng);
    }
    
    inline int in(double min,double max) { // outputs int
        // takes min and max of the range
        i = min; k = max;
        std::uniform_int_distribution<> dist(i,k);
        return dist(rng);
    }
    
};

// Normal Distribution
struct Norm {
    double i; // mean
    double k; // standard deviation
    RNG rng; // random number generator object
    Norm(RNG& rng_) : rng(rng_) {}
    // contructor takes unsigned int seed
    
    inline double doub(double mean,double stdev){ // outputs double
        // takes variables of the mean (i) and stadard deviation (k) for normal distribution
        i = mean; k = stdev;
        std::normal_distribution<double> set(i,k);
        return set(rng);
    }
    
    inline float flo(double mean,double stdev){ //outputs float
        // takes variables of the mean (i) and stadard deviation (k) for normal distribution
        i = mean; k = stdev;
        std::normal_distribution<float> set(i,k);
        return set(rng);
    }
};

/************************************************************************
 Creating random objects
 ************************************************************************/
// creating random object to be used in member functions
// Random numbers are generated from Mersenne Twister algorithm. See RandomWrap.hpp for more info
// Can replace with your own random number generator.

/////// Uniform distribution //////////
RNG mt = std::mt19937(25247638);
// creating random number generator object used to draw random numbers later
// number in parentheses is the seed. It can be changed and also be determined by random device
// RNG is typedef for random number generator defined in RandomWrap.hpp

Unif rUnif(mt); // object for unifrom distribution

/////////  Normal distrubution //////////
RNG mtNorm = std::mt19937(43755770);
// Need different seed for different distibutions so they aren't drawn from same random number list.

Norm rNorm(mtNorm); // object for normal distribution

/************************************************************************
 Defining function autocorrelation()
 ************************************************************************/
// Function for generation autocorrelated time series

void autocorrelated(double rho,float mean,float var,float series[],int length){
    // lambda - autocorrelation parameter
    // min - minimum value of sqaure wave
    // max - maximum value of square wave
    // length - length of series
    
    // float autocorrelated();
    
    float stdDev_ = sqrt(var);
    
    // setting first value
    series[0] = mean;
    
    // setting rest of series
    for (int i=1; i < length; i++){
        series[i] = mean + rho * (series[i-1] - mean) + (stdDev_ * sqrt(1 - pow(rho,2)) * rNorm.doub(0, 1));
    }
}


/************************************************************************
 Creating struct Agent
 ************************************************************************/
// Agents are the individuals in the model

struct Agent {
    ////////  Variables ////////
    double genotype[2][10];
    float phenotype;
    double env; // environmental component of phenotype
    bool alive; // Says whether or not the agent is alive
    int age;
    
    ////////  Functions ////////
    
    // constructor
    Agent(): alive(true) {}
    
    void setGenotype(float myArr[2][10], float e_){
        // setGenotype - sets the genotype and automatically calculates the phenotype
        // pars - myARR is an array of size 10 which represents genotpye
        //      - e_ is the environmental effect
        
        ////////// Setting genotype /////////
        for (int i=0; i < 2; i++){
            for (int k=0; k < 10; k++){
                genotype[i][k] = myArr[i][k];
            }
        }
        
        
        /////// Calculating phenotype ///////////
        phenotype = 0.0;
        
        // genetic effect
        for (int i=0;i < 2; i++){
            for (int k=0; k < 10; k++){
                phenotype += genotype[i][k];
            }
        }
        
        // random environmental effect
        phenotype += e_;
        
        // storing environmental effect
        env = e_;
    }
    
    void initate(){
        // initiate - gives initial values for genotype
        
        float arr[2][10];
        float choices[5][10];
        
        
        // Initial allelic options of each locus
        for (int i=0; i < 10; i++){ // which locus
            for(int k=0; k<5; k++){ // choosing 5 possible allelic values for each locus
                choices[k][i] = rNorm.doub(0,.01);
                // .01 = initial std of initial allelic values
            }
        }
        
        // Setting initial genotype
        for (int i=0; i < 2; i++){
            for (int k=0; k < 10; k++){
                arr[i][k] = choices[rUnif.in(0,4)][k];
            }
        }
        
        env = rNorm.doub(0,.01);
        
        setGenotype(arr,env); // random environment effect with SD .01
    }
    
    float getGenotype(){
        // outputs indiviudals genotype
        
        float g = 0.0;
        for (int i=0;i < 2; i++){
            for (int k=0; k < 10; k++){
                g += genotype[i][k];
            }
        }
        return(g);
    }
    
    void printInfo(){
        // printInfo - prints out genotype and phenotype
        // useful for testing
        
        std::cout << "Alive: " << alive << std::endl;
        
        std::cout << "Genotype: ";
        for (int i=0; i < 2; i++){
            std::cout << "H" << (i+1) << ": ";
            for (int k=0; k < 10; k++){
                std::cout << genotype[i][k] << ", ";
            }
            std::cout << std::endl;
        }
        
        std::cout << std::endl;
        
        std::cout << "Phenotype: " << phenotype << std::endl;
        
        std::cout << std::endl;
    }
    
};


/************************************************************************
 Creating class Sim
 ************************************************************************/

class Sim{
private:
    // ******* Parameters *******
    int K; // carrying capacity
    int B; // total reproductive effort
    float rc; // fraction of reproductive effort devoted to clonal repro
    float d; // cost of producing clonal offspring relative to sexual offspring
    float w; // strength of selection
    float ug; // gametic or sexual mutation rate
    float us; // somatic mutation rate
    float p; // correlation between parent and offspring env with clonal repro
    float rho; // autocorrelation parameter
    float mean_theta; // mean optimal trait value
    float var; // variance optimal trait value
    float surv; // probability of an adult surviving until the next generation
    
    // ******* Population vector *******
    std::vector <Agent> pop; // vector of Agents in the population
    std::vector <Agent> offspring; // vector for storing offspring
    std::vector <Agent> adults; // vector for adults that survived to next generation
    
public:
    
    float thetas[1000]; // time series of environmental values
    
    // ****  Functions *****
    void setup(int K_, int B_, float rc_, float d_, float p_, float w_,float ug_,float us_, float rho_,float mean_theta_,float var_, float surv_){
        // inputs all of the parameters
        
        K = K_;
        B = B_;
        rc = rc_;
        d = d_;
        p = p_;
        w = w_;
        ug = ug_;
        us = us_;
        rho = rho_;
        mean_theta = mean_theta_;
        var = var_;
        surv = surv_;
    }
    
    void init(int size){
        // init - creates the inital population and sets thetas
        // pars - initial size
        
        pop.clear(); // clears all previous adults
        adults.clear(); // clears all previous adults
        
        // sets individuals
        for (int i=0; i < size; i++){
            Agent A;
            A.initate();
            pop.push_back(A);
        }
        
        // sets theta series
        autocorrelated(rho, mean_theta, var, thetas, 1000);
    }
    
    void mate(Agent& mother, Agent& father){
        // random mating of of an agent
        // pars - Agent that is being cloned
        
        double Off = B*(rc/d) + B*(1-rc); // number of offspring
 
        for(int i=0; i < int(ceil(Off)); i++){
            double rn = rUnif.doub(0, 1); // random number to decide what happens
            
            
            // Probablity Clonal reproduction
            if(rn <= (rc/d)){ // if clonal reproduction
                Agent A; // creating offspring
                
                // Assigning genome
                float g[2][10];
                
                for (int j=0; j < 10; j ++){ // creating genome
                    g[0][j] = mother.genotype [0][j];
                    g[1][j] = mother.genotype [1][j];
                }
                
                /// mutation ///
                if(rUnif.doub(0,1) < us){ // if somatic mutation happens on 1st haplotype
                    int alleleM = rUnif.in(0,9);
                    g[0][alleleM] = g[0][alleleM] + rNorm.doub(0,.2236); // .2236 is sqrt(.05) variance
                }
                
                if(rUnif.doub(0,1) < us){ // if somatic mutation happens on 2nd haplotype
                    int alleleF = rUnif.in(0,9);
                    g[1][alleleF] = g[1][alleleF] + rNorm.doub(0,.2236); // .2236 is sqrt(.05) variance
                }
                
                // determining the environment effect
                double env_ = p * mother.env + sqrt(1 - (p*p)) * rNorm.doub(0, 1);
                
                // setting genotype
                A.setGenotype(g, env_);
                
                // adding to offspring population
                offspring.push_back(A);
            }
            
            // Probability Sexual reproduction
            else if ((rc/d) < rn <= ((rc/d)+(1-rc))) { // if sexual repoduction
                Agent A;
                
                // Assigning genome
                float g[2][10];
                
                for (int j=0; j < 10; j ++){ // creating mother haplotype
                    g[0][j] = mother.genotype [rUnif.in(0,1)][j];
                }
                
                for (int k=0; k < 10; k++){ // creating father haplotype
                    g[1][k] = father.genotype[rUnif.in(0,1)][k];
                }
                
                /// mutation ///
                if(rUnif.doub(0,1) < ug){ // if gametic mutation happens on mother haplotype
                    int alleleM = rUnif.in(0,9);
                    g[0][alleleM] = g[0][alleleM] + rNorm.doub(0,.2236); // .2236 is sqrt(.05) variance
                }
                
                if(rUnif.doub(0,1) < ug){ // if gametic mutation happens on father haplotype
                    int alleleF = rUnif.in(0,9);
                    g[1][alleleF] = g[1][alleleF] + rNorm.doub(0,.2236); // .2236 is sqrt(.05) variance
                }
                
                // determining the environmental effect
                double env_ = rNorm.doub(0,1);
                
                // setting genotype
                A.setGenotype(g, env_);
                
                // adding to offspring population
                offspring.push_back(A);
            }
            
        
            // no offspring produced with probability 1 - [(rc/d) + (1-rc)]
            // This only occurs if Off is not an integer
    
        }
    }
    
    
    void mateBurn(Agent& mother, Agent& father){ // mating for burnin(); all sexual
        // random mating of of an agent
        // pars - Agent that is being cloned
        
        for(int i=0; i < B; i++){
            Agent A;
            
            // Assigning genome
            float g[2][10];
            
            for (int j=0; j < 10; j ++){ // creating mother haplotype
                g[0][j] = mother.genotype [rUnif.in(0,1)][j];
            }
            
            for (int k=0; k < 10; k++){ // creating father haplotype
                g[1][k] = father.genotype[rUnif.in(0,1)][k];
            }
            
            /// mutation ///
            if(rUnif.doub(0,1) < ug){ // if mutation happens on mother haplotype
                int alleleM = rUnif.in(0,9);
                g[0][alleleM] = g[0][alleleM] + rNorm.doub(0,.2236); // .2236 is sqrt(.05) variance
            }
            
            if(rUnif.doub(0,1) < ug){ // if mutation happens on father haplotype
                int alleleF = rUnif.in(0,9);
                g[1][alleleF] = g[1][alleleF] + rNorm.doub(0,.2236); // .2236 is sqrt(.05) variance
            }
            
            // determining the environmental effect
            double env_ = rNorm.doub(0,1);
            
            // setting genotype
            A.setGenotype(g, env_);
            
            // adding to offspring population
            offspring.push_back(A);
        }
    }
    
    
    
    
    void step(float theta_){
        // runs a step of the simulation
        
        //******* selection *********
        int counter = 0;
        while (counter < pop.size()){
            // if dies erase from population
            if (rUnif.doub(0,1) > exp (- (pow((pop[counter].phenotype - theta_),2)) / (2*w))){
                pop.erase(pop.begin() + counter);
            }
            else{
                counter ++;
            }
        }
        
        //******* Reproduction *********
        // add adults that survived from last generation
        for(int i=0; i<adults.size(); i++){
            pop.push_back(adults[i]);
        }
        adults.clear();
        
        if(pop.size() < K){ // if less than K, all individuals mate as females
            for (int i = 0; i < pop.size(); i++){
                Agent father = pop[rUnif.in(0,(pop.size()-1))]; // randomly chosing a father
                mate(pop[i], father); // mating
                
                // Does parent survive until the next generation
                if(rUnif.doub(0,1) < surv){
                    adults.push_back(pop[i]);
                }
            }
        }
        else{ // if greater than K, only K individuals mate
            // randomly rearrange the order of individuals in the population vector
            std::shuffle(pop.begin(),pop.end(),mt);
            
            for (int i = 0; i < K; i++){ //if greater than K, only K individuals mate as female
                Agent father = pop[rUnif.in(0,(pop.size()-1))]; // randomly chosing a father
                mate(pop[i], father); // mating
                
                // Does parent survive until the next generation
                if(rUnif.doub(0,1) < surv){
                    adults.push_back(pop[i]);
                }
            }
        }
        
        //    //******* Updating population ********
        // killing off the adults and making offspring new adults in pop
        offspring.swap(pop); // swaps the contents of offspring and pop
        offspring.clear(); // clears the contents of offspring
    }
    
    void burnIN(){
        // burnIN - runs the population for 1000 generations with theta = 0
        // gets populations at mutation-selection balance and stable population structure
        
        for(int i=0; i < 1000; i++){
            //******* selection *********
            int counter = 0;
            while (counter < pop.size()){
                // if dies erase from population
                if (rUnif.doub(0,1) > exp (- (pow((pop[counter].phenotype - 0),2)) / (2*w))){
                    pop.erase(pop.begin() + counter);
                }
                else{
                    counter ++;
                }
            }
            
            
            if(pop.size() < K){ // if less than K, all individuals mate as females
                for (int i = 0; i < pop.size(); i++){
                    Agent father = pop[rUnif.in(0,(pop.size()-1))]; // randomly chosing a father
                    mateBurn(pop[i], father); // mating
                }
            }
            else{ // if greater than K, only K individuals mate
                // randomly rearrange the order of individuals in the population vector
                std::shuffle(pop.begin(),pop.end(),mt);
                
                for (int i = 0; i < K; i++){ //if greater than K, only K individuals mate as female
                    Agent father = pop[rUnif.in(0,(pop.size()-1))]; // randomly chosing a father
                    mateBurn(pop[i], father); // mating
                }
            }
            
            //    //******* Updating population ********
            // killing off the adults and making offspring new adults in pop
            offspring.swap(pop); // swaps the contents of offspring and pop
            offspring.clear(); // clears the contents of offspring

        }
    }
    
    void assay(){
        // series of mean theta at the end to see if population is actually adapted
        for(int i=0; i < 100; i++){
            step((mean_theta));
        }
    }
    
    void sim(int years){
        // initates and runs multiple steps of the simulation
        // pars - years: how many steps to run
        burnIN();
        for(int i=0; i < years; i++){
            //        std::cout << thetas[i] << ", " << popSize() << ", " << meanPhenotype() << std::endl;
            step(thetas[i]);
        }
        assay();
    
    }
    
    float meanPhenotype(){
        // outputs the mean phenotype of the population
        
        float pheno = 0;
        
        for (int i=0; i < pop.size(); i++){
            pheno += pop[i].phenotype;
        }
        
        return(pheno / pop.size());
    }
    
    unsigned long popSize(){
        // outputs population size
        return(pop.size());
    }
    
};


/************************************************************************
 Defining function loopThetas()
 ************************************************************************/
// Function that loops through multiple theta values

void loopThetas(int K_, int B_, float rc_, float d_, float p_, float w_, float ug_, float us_, float rho_, float var_, float surv_, float min, float max, float step, int reps){
    
    Sim test;
    int len = static_cast<int>((max - min)/step); // fix this and delete cout in sim
    
    std::cout <<  "Theta," << "P(Adapt)" << std::endl;
    
    for (int i=0; i <= len; i++){
        test.setup(K_, B_, rc_, d_, p_, w_, ug_, us_, rho_, (min + (i*step)), var_, surv_);
        float prob = 0; // tracking how many successes there are
        for (int j=0; j < reps; j++){
            test.init(K_ * B_);
            test.sim(1000); // change if you want a different length simulation
            
            if (test.popSize() > 100){
                prob++;
            }
        }
        prob = prob / reps; // calculates probability of adapation
        std::cout << (min + (i*step)) << "," << prob << std::endl;
    }
}





/************************************************************************
 Main Function
 ************************************************************************/

int main() {
    int K; int B; float rc; float d; float p; float w; float ug; float us; float rho; float var; float surv; float min; float max; float step; int reps;

    std::cout << "K, "; // carrying capacity
    std::cin >> K;
    std::cout << K << std::endl;
    std::cout << "B, "; // birth rate
    std::cin >> B;
    std::cout << B << std::endl;
    std::cout << "rc, "; // Average fraction of offspring that were produced clonally
    std::cin >> rc;
    std::cout << rc << std::endl;
    std::cout << "d, "; // cost of producing clonal offspring relative to sexual offspring
    std::cin >> d;
    std::cout << d << std::endl;
    std::cout << "p, "; // Association between the random components of each parent’s and its offspring’s phenotypes for clonal reproduction
    std::cin >> p;
    std::cout << p << std::endl;
    std::cout << "w, "; // strength of selection
    std::cin >> w;
    std::cout << w << std::endl;
    std::cout << "ug, "; // meiotic mutation rate
    std::cin >> ug;
    std::cout << ug << std::endl;
    std::cout << "us, "; // somatic mutation rate
    std::cin >> us;
    std::cout << us << std::endl;
    std::cout << "rho, "; // autocorrelation parameter
    std::cin >> rho;
    std::cout << rho << std::endl;
    std::cout << "var, "; // variance in theta
    std::cin >> var;
    std::cout << var << std::endl;
    std::cout << "surv, "; // probability of adults surviving
    std::cin >> surv;
    std::cout << surv << std::endl;
    std::cout << "min, "; // minimum value for looping
    std::cin >> min;
    std::cout << min << std::endl;
    std::cout << "max, "; // maximum value for looping
    std::cin >> max;
    std::cout << max << std::endl;
    std::cout << "step, "; // increment for looping
    std::cin >> step;
    std::cout << step << std::endl;
    std::cout << "reps, "; // number of runs
    std::cin >> reps;
    std::cout << reps << std::endl;
    std::cout << std::endl;

    // Record start time
    auto start = std::chrono::high_resolution_clock::now();

    loopThetas(K, B, rc, d, p, w, ug, us, rho, var, surv, min, max, step, reps); // imports


    // Outputs time the simulation
    auto finish = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = finish - start;
    std::cout << "Elapsed time: " << elapsed.count() << " s\n";
    
    return 0;
}
